js_typing <-
"var options = {
    strings: [\"%s\"],
    typeSpeed: 3,
    backSpeed: 1,
    cursorChar: ' ',
    loop: false
};
var typed = new Typed('#%s', options);"

shinyServer(function(input, output, session) {

    # Para atribuir classe CSS ao chat.
    shinyjs::removeCssClass(selector = "#MESSAGE_AREA",
                            class = "col-sm-4")
    shinyjs::addCssClass(selector = "#MESSAGE_AREA",
                         class = "col-lg-6")
    shinyjs::removeCssClass(selector = "#COLUMN_INPUT_MESSAGE",
                            class = "col-sm-6")
    shinyjs::addCssClass(selector = "#COLUMN_INPUT_MESSAGE",
                         class = "col-lg-6")
    # shinyjs::removeCssClass(selector = "#COLUMN_MENSAGENS_PRONTAS",
    #                         class = "col-sm-6")
    shinyjs::addCssClass(selector = "#COLUMN_MENSAGENS_PRONTAS",
                         class = "col-lg-6")

    # Objetos reativos -------------------------------------------------

    # Para incrementar o id de cada mensagem nova do chat.
    id_number <- reactiveVal(1)

    # Indica se o disclaimer foi exibido.
    disclaimer_displayed <- reactiveVal(FALSE)

    # Inicializa com a mensagem de boas-vindas.
    CHAT_HISTORY <- reactiveVal(initial_chat_history)
    # LAST_AGENT <- reactiveVal("ai")

    # Para que a caixa de texto do `user` apareça antes da caixa de
    # texto da `ai`.
    LAST_AGENT <- reactive({
        chat_history <- CHAT_HISTORY()
        tail(chat_history$role, n = 1L)
    })
    LAST_AGENT_DELAY <- LAST_AGENT |>
        debounce(millis = 1000)

    # Autenticação -----------------------------------------------------

    # Para armaazenar o token de autenticação.
    tokenSession <- reactiveVal()

    # Pega os parâmetros da URL.-
    url_params <- reactive({
        # list(token = "teste")
        parseQueryString(session$clientData$url_search)
    })

    # Escolhe a interface principal -----------------------------------

    observeEvent(url_params(), {
        token <- url_params()$token
        if (is.null(token)) {
            print("Token vazio.")
            token_session <- list(id_session = NULL, status = 401L)
        } else {
            print("Token aplicado.")
            # dput(check_agritoken("teste"))
            # token_session <- list(id_session = 63L, status = 200L, id_user = 3191L)
            token_session <- check_agritoken(token)
        }

        if (token_session$status == 200) {
            # Atualiza o reativo com o as informações de
            # autenticação.
            tokenSession(token_session)
            print("Token válido.")
        } else {
            print("Token inválido.")
            chat_history <-
                wb$create_chat_history(prompt = welcome_stranger,
                                       role = "ai")
            CHAT_HISTORY(chat_history)
            shinyjs::disable("SEND_MESSAGE")
            shinyjs::disable("INPUT_MESSAGE")

        }

    })

    # output$TOKEN <- renderPrint({
    #     list(url_params(),
    #          tokenSession(),
    #          Sys.time())
    # })

    # UI do chat -------------------------------------------------------

    observeEvent(CHAT_HISTORY(), {

        # Gera o id para a nova mensagem.
        id <- paste0("typing-text-", id_number()) # id=1

        # chat_history <- tail(initial_chat_history, n = 1)
        chat_history <- CHAT_HISTORY() |>
            tail(n = 1)

        # IMPORTANT: Salva a mensagem no banco de dados.
        id_session <- tokenSession()$id_session
        if (is.null(id_session) || length(id_session) == 0L) {
            id_session <- -999
        }
        db_save_dt(db_table =  "gg_hist_chat",
                   data_table = data.frame(
                       fk_session = id_session,
                       message = chat_history$prompt,
                       agente = chat_history$role)
                   )

        role <- chat_history[["role"]]
        rs <- chat_history[["prompt"]] |>
            repair_string()

        if (role == "ai") {
            L <- chat_message(role = role,
                              chat_metadata = chat_metadata) |>
                append(value = list(text = tags$div(id = id)))
        } else {
            L <- chat_message(role = role,
                              chat_metadata = chat_metadata) |>
                append(value = list(text = rs))
        }

        # Adiciona a mensagem ao chat.
        updateUserMessages(
            session = session,
            id = "MESSAGE_AREA",
            action = "add",
            content = list(
                # author = L$author,
                author = "",
                # date = L$date,
                date = "",
                image = L$image,
                type = L$type,
                text = L$text
            )
        )

        if (role == "ai") {
            # Aplica o efeito de digitação na mensagem.
            shinyjs::runjs(sprintf(js_typing, rs, id))
        }

        # Exibe a mensagem no console.
        if (getOption("test_mode", FALSE)) {
            # cat(id, "\n")
            cat("----------------------------------------------------------------------\n\n")
            cat("  Autor: ", L$author, "\n")
            cat("  Mensagem:\n\n")
            rs |>
                strwrap(width = 64, indent = 4, exdent = 4) |>
                cat(sep = "\n", "")
        }

        # Incrementa o id para a próxima mensagem.
        id_number(id_number() + 1)
    })

    # Disclaimer -------------------------------------------------------

    disclaimer_modal <- function() {
        modalDialog(
            title = "Termos de uso",
            includeMarkdown("disclaimer.md"),
            footer = tagList(
                checkboxInput(
                    inputId = "ACCEPT_TERMS",
                    label = "Eu li e concordo com os termos de uso."
                ),
                modalButton("Prosseguir")
            )
        )
    }

    observeEvent(tokenSession(), {
        id_user <- tokenSession()$id_user
        # cat("Usuário: ", id_user)
        user_accepted <-
            terms_accepted(id_user)
        if (isTRUE(user_accepted)) {
            updateCheckboxInput(session, "ACCEPT_TERMS", value = TRUE)
            disclaimer_displayed(TRUE)
            cat("Termos já foram aceitos pelo usuário.")
        } else {
            cat("Termos ainda não foram aceitos pelo usuário.")
        }
    })

    observeEvent(input$ACCEPT_TERMS, {
        id_user <- tokenSession()$id_user
        # cat("Usuário: ", id_user, "\n")
        if (!is.null(id_user)) {
            user_accepted <-
                terms_accepted(id_user)
            if (input$ACCEPT_TERMS && !isTRUE(user_accepted)) {
                cat("Termos acabam de ser aceitos pelo usuário.")
                db_comand(sprintf("INSERT INTO gg_term (fk_user) VALUES (%s)",
                                  id_user))
                disclaimer_displayed(TRUE)
            }
        }
    })

    # Submissão de uma pergunta ----------------------------------------

    observeEvent(input$SEND_MESSAGE, {
        req(input$INPUT_MESSAGE)

        # Exibe o disclaimer.
        if (!disclaimer_displayed()) {
            showModal(disclaimer_modal())
            disclaimer_displayed(TRUE)
        } else {
            if (nchar(input$INPUT_MESSAGE) > 10) {
                # Atualiza o chat.
                chat_history <- CHAT_HISTORY()
                chat_history <- wb$update_chat_history(
                    prompt = input$INPUT_MESSAGE,
                    role = "user",
                    chat_history = chat_history
                )
                CHAT_HISTORY(chat_history)

                # Remove o texto.
                updateTextAreaInput(
                    session = session,
                    inputId = "INPUT_MESSAGE",
                    value = "")
            }
        }
    })

    # Consulta a API ---------------------------------------------------

    observeEvent(LAST_AGENT_DELAY(), {
        chat_history <- CHAT_HISTORY()
        # Último agente a escrever no chat.
        last_agent <- tail(chat_history$role, n = 1L)
        if (last_agent == "user") {
            # Última mensagem.
            last_input <- tail(chat_history$prompt, n = 1L)

            # Mostra o sinal de carregamento.
            shinyjs::show("loading")

            if (getOption("test_mode", TRUE)) {
                answer <-
                    c("Modo de teste está ativo. Esse é um texto aleatório do pacote `fortunes`.\n",
                      fortunes::fortune()[["quote"]]) |>
                    paste(collapse = "\n")
                answer <- paste0(answer, "<br><br><a href='https://www.google.com'>Link</a>")
                # cat(answer, "\n")
                chat_history <- wb$update_chat_history(
                    prompt = answer,
                    role = "ai",
                    chat_history = chat_history
                )
                Sys.sleep(5)
                resp <- list(chat_history = chat_history)
            } else {
                # Faz a requisição.
                resp <- wb$grao_guru(chat_history = chat_history,
                                     vector_store = vector_store,
                                     token = Sys.getenv("OPENAI_API"))
                resp <- adds_links(resp)
            }

            # Atualiza o chat.
            CHAT_HISTORY(resp$chat_history)

            # Esconde o sinal de carregamento.
            shinyjs::hide("loading")
        }
    })

})
